/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include <fstream>
#include <string>
#include <glib.h>
#include <pan/general/debug.h>
#include <pan/general/foreach.h>
#include "ui-prefs.h"

using namespace pan;

namespace
{
  struct MyContext {
    UI_Prefs& prefs;
    std::string text;
    std::string name;
    int x, y, w, h;
    MyContext (UI_Prefs& p): prefs(p), x(0), y(0), w(0), h(0) {}
  };

  // called for open tags <foo bar='baz'>
  void start_element (GMarkupParseContext *context,
                      const gchar         *element_name,
                      const gchar        **attribute_names,
                      const gchar        **attribute_vals,
                      gpointer             user_data,
                      GError             **error)
  {
    const std::string s (element_name);
    MyContext& mc (*static_cast<MyContext*>(user_data));
    mc.text.clear ();

    if (s == "geometry") {
      const char * name(0);
      int x(0), y(0), w(0), h(0);
      for (const char **k(attribute_names), **v(attribute_vals); *k; ++k, ++v)
        if (!strcmp (*k,"name"))  name = *v;
        else if (!strcmp(*k,"x")) x = atoi (*v);
        else if (!strcmp(*k,"y")) y = atoi (*v);
        else if (!strcmp(*k,"w")) w = atoi (*v);
	else if (!strcmp(*k,"h")) h = atoi (*v);
      if (name && *name)
        mc.prefs.set_geometry (name, x, y, w, h);
    }

    if (s == "flag") {
      const char * name (0);
      bool b (false);
      for (const char **k(attribute_names), **v(attribute_vals); *k; ++k, ++v)
        if (!strcmp (*k,"name"))  name = *v;
        else if (!strcmp(*k,"value")) b = *v && **v && tolower(**v)=='t';
      if (name && *name)
        mc.prefs.set_flag (name, b);
    }
  }

  // Called for close tags </foo>
  void end_element    (GMarkupParseContext *context,
                       const gchar         *element_name,
                       gpointer             user_data,
                       GError             **error)
  {
  }

  void text (GMarkupParseContext *context,
             const gchar         *text,
             gsize                text_len,  
             gpointer             user_data,
             GError             **error)
  {
    static_cast<MyContext*>(user_data)->text.assign (text, text_len);
  }
}

void
UI_Prefs :: set_from_xml (const StringView& xml)
{
  MyContext mc (*this);
  GMarkupParser p;
  p.start_element = start_element;
  p.end_element = end_element;
  p.text = text;
  p.passthrough = 0;
  p.error = 0;
  GMarkupParseContext* c (
    g_markup_parse_context_new (&p, (GMarkupParseFlags)0, &mc, 0));
  GError * gerr (0);
  if (g_markup_parse_context_parse (c, xml.str, xml.len, &gerr)) {
    // FIXME
  }
  g_markup_parse_context_free (c);
}

void
UI_Prefs :: set_from_file (const StringView& filename)
{
  gchar * txt (0);
  gsize len (0);
  if (g_file_get_contents (filename.to_string().c_str(), &txt, &len, 0))
    set_from_xml (StringView(txt,len));
  g_free (txt);
}

namespace
{
  const int indent_char_len (2);

  std::string indent (int depth) { return std::string(depth*indent_char_len, ' '); }

  std::string escaped (const std::string& s)
  {
    char * pch = g_markup_escape_text (s.c_str(), s.size());
    const std::string ret (pch);
    g_free (pch);
    return ret;
  }

  std::string escaped (const bool b)
  {
    return b ? "true" : "false";
  }
}

void
UI_Prefs :: serialize (std::ostream&  out) const
{
  int depth (0);
  out << "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n";
  out << indent(depth++) << "<preferences>\n";

  foreach_const (flags_t, _flags, it)
    out << indent(depth) << "<flag name='" << it->first << "' value='" << (it->second ? "true" : "false") << "'/>\n";

  foreach_const (window_to_geometry_t, _window_to_geometry, it) {
    if (it->second.x || it->second.y || it->second.w || it->second.h) {
      out << indent(depth) << "<geometry name='" << it->first << "'";
      if (it->second.x) out << " x='" << it->second.x << "'";
      if (it->second.y) out << " y='" << it->second.y << "'";
      if (it->second.w) out << " w='" << it->second.w << "'";
      if (it->second.h) out << " h='" << it->second.h << "'";
      out << "/>\n";
    }
  }

  out << indent(--depth) << "</preferences>\n";
}

UI_Prefs :: UI_Prefs (const StringView& filename):
  _filename (filename.to_string())
{
  set_from_file (_filename);
}

UI_Prefs :: ~UI_Prefs ()
{
  save ();
}

void 
UI_Prefs :: save ()
{
  std::ofstream out (_filename.c_str());
  serialize (out);
  out.close ();
}

/**
***
**/

bool
UI_Prefs :: get_geometry (const StringView& window_name, int& x, int& y, int& w, int& h) const
{
  window_to_geometry_t::const_iterator it (_window_to_geometry.find (window_name));
  if (it != _window_to_geometry.end()) {
    const Geometry& g (it->second);
    x = g.x;
    y = g.y;
    w = g.w;
    h = g.h;
  }
  return it != _window_to_geometry.end();
}

void
UI_Prefs :: set_geometry (const StringView& key, int x, int y, int w, int h)
{
  Geometry& g (_window_to_geometry[key]);
  g.x = x;
  g.y = y;
  g.w = w;
  g.h = h;
}

void
UI_Prefs :: set_default_geometry (const StringView& key, int x, int y, int w, int h)
{
  if (!_window_to_geometry.count (key))
    set_geometry (key, x, y, w, h);
}

#define UI_PREFS_WIDGET_KEY "ui-prefs-widget-key"

void
UI_Prefs :: size_allocated_cb (GtkWidget      * widget,
                               GtkAllocation  * alloc,
                               gpointer         pointer)
{
  UI_Prefs& self (*static_cast<UI_Prefs*>(pointer));
  const char * key ((const char*) g_object_get_data (G_OBJECT(widget), UI_PREFS_WIDGET_KEY));

  int x(0), y(0);
  if (GTK_IS_WINDOW (widget))
    gtk_window_get_position (GTK_WINDOW(widget), &x, &y);
  if (GTK_IS_WINDOW (widget) || self._sentinels[key]->save_widget_allocations_to_prefs())
    self.set_geometry (key, x, y, alloc->width, alloc->height);
}

void
UI_Prefs :: set_window (const StringView& key, GtkWindow* window, int x, int y, int w, int h)
{
  get_geometry (key, x, y, w, h);
  gtk_window_move (window, x, y);
  gtk_window_set_default_size (window, w, h);
  g_object_set_data_full (G_OBJECT(window), UI_PREFS_WIDGET_KEY, g_strdup(key.to_string().c_str()), g_free);
  g_signal_connect (window, "size-allocate", G_CALLBACK(size_allocated_cb), this);
}

void
UI_Prefs :: set_widget (const StringView& key, GtkWidget* widget,
                        SaveWidgetAllocationsToPrefs* sentinel,
                        int w, int h)
{
  _sentinels[key] = sentinel;

  int unused_x, unused_y;
  get_geometry (key, unused_x, unused_y, w, h);
  gtk_widget_set_size_request (widget, w, h);
  g_object_set_data_full (G_OBJECT(widget), UI_PREFS_WIDGET_KEY, g_strdup(key.to_string().c_str()), g_free);
  g_signal_connect (widget, "size-allocate", G_CALLBACK(size_allocated_cb), this);
}

bool
UI_Prefs :: get_flag (const StringView& key, bool fallback) const
{
  if (!_flags.count (key))
    _flags[key] = fallback;
  return _flags[key];
}

void
UI_Prefs :: set_flag (const StringView& key, bool value)
{
  _flags[key] = value;
}
